#include <oxstd.h>
#include <oxprob.h>
#include <oxdraw.h>
#include <oxfloat.h>
#import "maximize"
//#import <packages/maxsa/maxsa>

static decl s_mScores, s_mScaledScores, s_mAB, s_mYr, s_mFactors;
static decl s_iNumObs, s_iNumFactors, s_iRestrict;
static decl s_vFactors, s_iPrint;
static decl s_vParmsOrig;

restrictParms(const vParmsBegin)
{
	decl vParms = vParmsBegin .* s_vParmsOrig;
	decl vRestrictedParms = vParms;

		// A parameters
		vRestrictedParms[0] = vParms[0];
		// B parameters
		vRestrictedParms[1] = 1 ./ (1 + exp(-vParms[1]));
		// beta
		vRestrictedParms[2] = exp(vParms[2]);
		// vZr
		vRestrictedParms[3] =  vParms[3];
	
	return vRestrictedParms;
}

getStartingValues()
{
	decl vParms = 0;

	
		s_iNumFactors = 1;
	
		vParms = 0.07534 | 
		         3.47959 |
				 log(2.6144) | 0.34975 ;
	
	return vParms;
}

getGasMatrices(const vRestrictedParms, const pmA, const pmB, const pdBetaR, const pvZr, const pmZr, const iPrint)
{
	decl inc = 0, temp;

	 	pmA[0] = vRestrictedParms[0];
		pmB[0] = vRestrictedParms[1];
		pdBetaR[0] = vRestrictedParms[2];
		pvZr[0] = vRestrictedParms[3];
		pmZr[0] = 1;
}

logLikelihood(const vParms, const pdLoglike, const pvG, const pmHess)
{

	decl mAB, mA, mB, mScores, mInfo, mInvInfo, mFactors, mYr;
	decl vYr, vRR, vAB, vBB, vZr, mZr;
	decl vFactors, dMu, dAlpha, dBeta, dBetaR, dLoglike, vScore, vSt, vRestrictedParms;
	decl iNumObs, iNumFactors;
	decl mScaledScores, mEigval, mEigvec, mUU;
	decl temp, t, i, j, k, sign, dA, dB;
	pdLoglike[0] = -.Inf;					
			
	iNumFactors = 1;
	mYr = s_mYr;
	iNumObs = columns(mYr);
							 
	//vRestrictedParms = restrictParms(vParms .* s_vParmsOrig);
	if(s_iRestrict==1)
	{
		vRestrictedParms = restrictParms(vParms);
	}
	else
	{
		vRestrictedParms = vParms;
	}
	getGasMatrices(vRestrictedParms, &mA, &mB, &dBetaR, &vZr, &mZr, s_iPrint);

	mFactors = zeros(iNumFactors, iNumObs);
	mScores = zeros(iNumFactors, iNumObs);
	mScaledScores = zeros(iNumFactors, iNumObs);
	vFactors = zeros(iNumFactors,1);

	mAB = zeros(2,iNumObs);
	
	vBB = 1 ~ -1;
	
	dLoglike = 0;
	
	// then the rest of the observations
	for(t=0;t<iNumObs;t++)
	{

		// initialize score and info matrix
		mFactors[][t] = vFactors;							
		 
 		dMu = exp(vZr + mZr*vFactors)/(1 + exp(vZr + mZr*vFactors));
		
		//dSigma2R = dMu*(1 - dMu)/(1 + dBetaR);
		dAlpha = dBetaR*dMu; 
		dBeta = dBetaR*(1 - dMu);

		mAB[0][t] = dAlpha;
		mAB[1][t] = dBeta;

		
		vYr = mYr[t];

	//	println(t~dMu~dAlpha~dBeta~vYr);
		
		dLoglike += sumc( loggamma(dAlpha + dBeta) - loggamma(dAlpha) - loggamma(dBeta) +
						(dAlpha - 1)*log(vYr) + (dBeta - 1)*log(1 - vYr) );
					
		vAB = sumc(
				polygamma(dAlpha + dBeta,0) +
				(
					log(vYr) - polygamma(dAlpha,0)
					~
         			log(1 - vYr) - polygamma(dBeta,0)
				)
			)';	  
		vScore = dBetaR*dMu*(1-dMu)*mZr'*vBB*vAB;
		temp = diag(polygamma(dAlpha,1) | polygamma(dBeta,1)) - polygamma(dAlpha + dBeta,1);
		//mInfo = sqr(dBetaR*dMu*(1-dMu))*mZr'(rows(vYr)*(vBB*temp*vBB'))*mZr;
		mInfo = sqr(dBetaR*dMu*(1-dMu))*mZr'((vBB*temp*vBB'))*mZr;
	
			
		// update with GAS step
		//eigensym(mInfo,&mEigval,&mEigvec);
		//mUU = mEigvec[][vIndex];
		//mInvInfo = mUU*(diag(1 ./mEigval[vIndex]))*mUU';
		mInvInfo = sqrt(1/mInfo);
		
		vSt = mInvInfo * vScore;
		mScores[][t] = vScore;
		mScaledScores[][t] = vSt;
		
		vFactors = mB*vFactors + mA * vSt;
	}
					  
	s_mScores = mScores;
	s_mScaledScores = mScaledScores;
	s_vFactors = vFactors;

	s_mAB = mAB;
	s_mFactors = mFactors;
	

	pdLoglike[0] = double(dLoglike);
	
	if(isnan(pdLoglike[0]))
	{
		return 0;
	}
	else
	{
		return 1;
	}
}

computeGradient(const func, const vParms, const deltagrad)
{
	decl m, k, mI, mGradient, fbase, vParms2, vParmsTemp;
  	m = rows(vParms);
	k = columns(vParms);
	vParmsTemp = vParms;
	if(m==1)
	{
		vParmsTemp = vParms'; 
	}
	m = rows(vParmsTemp);
	mI = unit(m);
	fbase = func(vParmsTemp);
	mGradient = zeros(m,m);
	for(decl i=0;i<m;i++)
	{
		vParms2 = vParmsTemp + deltagrad*mI[][i];
		mGradient[][i] = (func(vParms2)-fbase)/deltagrad ;
	}
	return mGradient;
}

standardErrors(const objFunction, const restrictionFunction, const vParms, const pvParms, const pvStdErrors)
{
	decl mHess, mJacobian, ir;

	// ============================================ //
	//	compute standard errors
	// ============================================ //
	ir = Num2Derivative(objFunction, vParms, &mHess);
    if (!ir)
    {	print("Covar() failed in numerical second derivatives\n");
		pvParms[0] = zeros(vParms);
		pvStdErrors[0] = zeros(vParms); 
        return;
    }
    mJacobian = computeGradient(restrictionFunction,vParms, 0.001);
	pvParms[0] = restrictionFunction(vParms);
	pvStdErrors[0] = sqrt(diagonal(mJacobian*(invert(-mHess))*mJacobian'))';
}

simulateBetaModel(const iNumObs, const pmYr, const pvParmsTrue)
{

	decl mAB, mA, mB, mInfo, mInvInfo, mFactors, mYr;
	decl vYr, vProbs, vRR, vAB, vBB, vZr, mZr;
	decl vFactors, dMu, dAlpha, dBeta, dBetaR, dLoglike, vScore, vSt;
	decl iNumFactors;
	decl mEigval, mEigvec, mUU;
	decl temp, t, i, j, k, sign, dA, dB;
			
	iNumFactors = 1;
	mYr = zeros(1,iNumObs);
							 
	mA = 0.02;
	mB = 0.985;
	dBetaR = 3.25;
	vZr = 0.2;
	mZr = 1;

	pvParmsTrue[0] = mA | mB | dBetaR | vZr ;  
	
	mFactors = zeros(iNumFactors, iNumObs);
	
	mAB = zeros(2,iNumObs);
	
	vFactors = zeros(iNumFactors,1);

	vBB = 1 ~ -1;
	
	dLoglike = 0;
	
	// then the rest of the observations
	for(t=0;t<iNumObs;t++)
	{

		// initialize score and info matrix
		mFactors[][t] = vFactors;							
		 
 		dMu = exp(vZr + mZr*vFactors)/(1 + exp(vZr + mZr*vFactors));
		
		dAlpha = dBetaR*dMu; 
		dBeta = dBetaR*(1 - dMu);
		
		mAB[0][t] = dAlpha;
		mAB[1][t] = dBeta;

		vYr = ranbeta(1,1,dAlpha,dBeta);
		mYr[t] = vYr;
		
		vAB = sumc(
				polygamma(dAlpha + dBeta,0) +
				(
					log(vYr) - polygamma(dAlpha,0)
					~
         			log(1 - vYr) - polygamma(dBeta,0)
				)
			)';	  
		vScore = dBetaR*dMu*(1-dMu)*mZr'*vBB*vAB;
		temp = diag(polygamma(dAlpha,1) | polygamma(dBeta,1)) - polygamma(dAlpha + dBeta,1);
		
		//mInfo = sqr(dBetaR*dMu*(1-dMu))*mZr'(rows(vYr)*(vBB*temp*vBB'))*mZr;
		mInfo = sqr(dBetaR*dMu*(1-dMu))*mZr'((vBB*temp*vBB'))*mZr;

		mInvInfo = sqrt(1/mInfo);
		
		vSt = mInvInfo * vScore;
		
		vFactors = mB*vFactors + mA * vSt;
	}
	pmYr[0] = mYr;
					 
}																							 

main()
{

	decl vParms, vParmsTrue, vParmsMLE, vStdErrors, vGrad, mHess;
	decl dLoglike, ir, iNumObs;
	decl mYr;

	iNumObs = 1000;
	simulateBetaModel(iNumObs, &mYr, &vParmsTrue);

	s_mYr = mYr;

	s_iNumObs = iNumObs;
	s_iRestrict = 1;
	s_iPrint = 0;
	
	s_vParmsOrig = 1;
	vParms = getStartingValues();

	logLikelihood(vParms, &dLoglike, &vGrad, &mHess);

	//println(dLoglike);
	//return;
	
	MaxControl(-1,1,0);
	ir = MaxBFGS(logLikelihood, &vParms, &dLoglike, 0, 1);

	s_vParmsOrig = fabs(vParms);
	s_vParmsOrig = exp(round( 10* log(s_vParmsOrig) ) / 10 );
	vParms = 2 * (vParms .> 0) - 1;
	ir = MaxBFGS(logLikelihood, &vParms, &dLoglike, 0, 1);

	println("**********************************************************************************************");
	println("**********************************************************************************************");
	println("**********************************************************************************************");
	println("\n", MaxConvergenceMsg(ir),
          " using numerical derivatives",
          "\nFunction value = ", dLoglike);
	println("loglikelihood: ",dLoglike);	  

	println(restrictParms(vParms));
	
	vParms .*= s_vParmsOrig;
	s_vParmsOrig = 1;
	  

	println(" ");
	for(decl j=0;j<rows(vParms); j++)
		print(vParms[j]," | ");
	println(" ");
				
	standardErrors(logLikelihood, restrictParms, vParms, &vParmsMLE, &vStdErrors);
					  
	s_iPrint = 1;
	s_iRestrict = 0;
	logLikelihood(vParmsMLE,&dLoglike,0,0);
	logLikelihood(vStdErrors,&dLoglike,0,0);

	println(vParmsTrue~vParmsMLE~vStdErrors);

}