#include <oxstd.h>
#include <oxprob.h>
#include <oxdraw.h>
#include <oxfloat.h>
#import "maximize"
//#import <packages/maxsa/maxsa>

static decl s_mYg, s_mFactors;
static decl s_iNumObs, s_iNumFactors, s_iRestrict;
static decl s_iNumMacroFactors, s_iNumMacros;
static decl s_vFactors, s_iPrint, s_mZg;
static decl s_vParmsOrig;

restrictParms(const vParmsBegin)
{
	decl vParms = vParmsBegin .* s_vParmsOrig;
	decl vRestrictedParms = vParms;
	decl temp;
	decl inc = 0;

		// A parameters
		vRestrictedParms[inc : inc + s_iNumFactors - 1] = vParms[inc : inc + s_iNumFactors - 1];
		inc += s_iNumFactors;
		// B parameters
		vRestrictedParms[inc : inc + s_iNumFactors - 1] = 1 ./ (1 + exp(-vParms[inc : inc + s_iNumFactors - 1]));
		inc += s_iNumFactors;
		// mZg
		inc += (s_iNumMacros - s_iNumMacroFactors)*s_iNumMacroFactors;
		// vSigma2 parameters for macros
		vRestrictedParms[inc: inc + s_iNumMacros - 1] = exp(vParms[inc: inc + s_iNumMacros - 1]);
		
	return vRestrictedParms;
}

getStartingValues()
{
	decl vParms = 0;
	   
	    s_iNumMacros = rows(s_mYg);
		s_iNumFactors = s_iNumMacroFactors = 3;
		
	vParms = 0.005986 | 0.007311 | 0.00332 |
	         3.39919 | 3.80618 | 3.06523 |
	       -0.1 | -0.1 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 |
	-2.918 | -1.67504 | -0.612246 | 1.12668 | 1.11497 | 1.45307 ; 
	
	return vParms;
}

getGasMatrices(const vRestrictedParms, const pmA, const pmB, const pmZg,
					const pvSigma, const iPrint)
{
	decl inc = 0, temp;

	pmZg[0] = 0;
	pvSigma[0] = 0;
		

	 	pmA[0] = diag(vRestrictedParms[inc:inc + s_iNumFactors - 1]);
		inc += s_iNumFactors;
		pmB[0] = diag(vRestrictedParms[inc:inc + s_iNumFactors - 1]);
		inc += s_iNumFactors;
		pmZg[0] = 1 ~ 0 ~ 0 |
				  vRestrictedParms[inc++] ~ 1 ~ 0 |
				  vRestrictedParms[inc++] ~ vRestrictedParms[inc++] ~ 1 |
				  vRestrictedParms[inc++] ~ vRestrictedParms[inc++] ~ vRestrictedParms[inc++] |
				  vRestrictedParms[inc++] ~ vRestrictedParms[inc++] ~ vRestrictedParms[inc++] |
				  vRestrictedParms[inc++] ~ vRestrictedParms[inc++] ~ vRestrictedParms[inc++] ;
		pvSigma[0] = vRestrictedParms[inc:inc + s_iNumMacros - 1] ;
	
}

logLikelihood(const vParms, const pdLoglike, const pvG, const pmHess)
{

	decl mA, mB, mInfo, mInvInfo, mFactors, mYg;
	decl vZg, mZg, vSigma2, mSigma, vMu, vEps;
	decl vIndex, vFactors, dLoglike, vScore, vSt, vRestrictedParms;
	decl iWarning, iNumObs, iNumFactors, iNumMacros, mIndex;
	decl mEigval, mEigvec, mUU;
	decl t, i, j, k, sign, dA, dB;

	pdLoglike[0] = -.Inf;					
			
	iNumObs = s_iNumObs;
	iNumFactors = s_iNumFactors;
	iNumMacros = s_iNumMacros;
	mYg = s_mYg;

	if(s_iRestrict==1)
	{
		vRestrictedParms = restrictParms(vParms);
	}
	else
	{
		vRestrictedParms = vParms;
	}
	getGasMatrices(vRestrictedParms, &mA, &mB, &mZg, &vSigma2, s_iPrint);

	mSigma = diag(vSigma2);
			
	mFactors = zeros(iNumFactors, iNumObs);

	vFactors = zeros(iNumFactors,1);

	dLoglike = 0;

	if(iNumMacros>0)
	{
		iWarning = oxwarning(0);
		oxwarning(iWarning);
	}
	else
	{
		mYg = ones(1,iNumObs)*.NaN;
	}
	dLoglike = 0;

	mIndex = zeros(iNumFactors,iNumFactors);

	// then the rest of the observations


	for(t=0;t<iNumObs;t++)
	{

		// initialize score and info matrix
		vScore = zeros(iNumFactors,1);
		mInfo = zeros(iNumFactors,iNumFactors);
		mFactors[][t] = vFactors;							
		
		if(iNumMacros>0)
		{
			vMu = mZg * vFactors;
			vEps = mYg[][t] - vMu;
 
			// update: Gaussian part of the likelihood

			dLoglike += sumc( isdotnan(mYg[][t]) .? 0 .: (-0.5*log(M_2PI) -0.5*log(vSigma2) - 0.5 .* sqr(vEps) ./ vSigma2) );

			vScore = mZg' * ( isdotnan(mYg[][t]) .? 0 .: (vEps ./vSigma2));
			mInfo = mZg' * ( diag(isdotnan(mYg[][t]) .? 0 .: (1 ./vSigma2) )) * mZg;
		}
					 
		// check likelihood existence
		if(isnan(dLoglike))
		{
			break;
		}

		// update with GAS step
			if(isnan(mInfo) || (sumc(vecr(isdotinf(mInfo))) != 0))
			{
				mInvInfo = unit(s_iNumFactors);
			}
			else
			{
				eigensym(mInfo,&mEigval,&mEigvec);
			
				vIndex = <>;
				for(j=0;j<iNumFactors;j++)
				{
					if(mEigval[j]>0)
					{
						vIndex |= j;
					}
				}
				if(vIndex==<>)
				{
					mInvInfo = zeros(s_iNumFactors);
				}
				else
				{
					mUU = mEigvec[][vIndex];
					mInvInfo = mUU*(diag(1 ./sqrt(mEigval[vIndex])))*mUU';
				}
			}
	
		vSt = mInvInfo * vScore;
		
		vFactors = mB*vFactors + mA * vSt;
	}
					  
	s_vFactors = vFactors;
	s_mFactors = mFactors;
	
	pdLoglike[0] = double(dLoglike);

	if(isnan(pdLoglike[0]))
	{
		return 0;
	}
	else
	{
		return 1;
	}
}

computeGradient(const func, const vParms, const deltagrad)
{
	decl m, k, mI, mGradient, fbase, vParms2, vParmsTemp;
  	m = rows(vParms);
	k = columns(vParms);
	vParmsTemp = vParms;
	if(m==1)
	{
		vParmsTemp = vParms'; 
	}
	m = rows(vParmsTemp);
	mI = unit(m);
	fbase = func(vParmsTemp);
	mGradient = zeros(m,m);
	for(decl i=0;i<m;i++)
	{
		vParms2 = vParmsTemp + deltagrad*mI[][i];
		mGradient[][i] = (func(vParms2)-fbase)/deltagrad ;
	}
	return mGradient;
}

standardErrors(const objFunction, const restrictionFunction, const vParms, const pvParms, const pvStdErrors)
{
	decl mHess, mJacobian, ir;

	// ============================================ //
	//	compute standard errors
	// ============================================ //
	ir = Num2Derivative(objFunction, vParms, &mHess);
    if (!ir)
    {	print("Covar() failed in numerical second derivatives\n");
		pvParms[0] = zeros(vParms);
		pvStdErrors[0] = zeros(vParms); 
        return;
    }
    mJacobian = computeGradient(restrictionFunction,vParms, 0.001);
	pvParms[0] = restrictionFunction(vParms);
	pvStdErrors[0] = sqrt(diagonal(mJacobian*(invert(-mHess))*mJacobian'))';
}

getData(const pmYg, const pvTime)
{
	decl iNumGrades, mTransitions, mExposures, vSelect, iNumObs, temp, mWithdrawals, mEntrances;
	pmYg[0] = loadmat("CSKL_mMacrosMonthly.in7");
	// 1. ind pro
	// 2. unemployment
	// 3. term spread
	// 4. credit spread
	// 5. stock market return
	// 6. realized vol
	// 7. corporate net cash flows
	// 8. real gdp
	// 9. corporate profits
	vSelect = <0;1;7;3;4;5>;
	pmYg[0] = pmYg[0][vSelect][];
	pvTime[0] = loadmat("CSKL_vTimeMonthly.in7");
}
																									 

main()
{

	decl vTime, vParms, vParmsMLE, vStdErrors, vDraws, vGrad, mHess;
	decl dLoglike, ir, t;
	decl mYg;		
	
	getData(&mYg, &vTime);
				 
	s_mYg = mYg;

	s_mYg = (s_mYg - meanr(deletec(s_mYg))) ./sqrt(varr(deletec(s_mYg)));
	s_mYg = s_mYg[0:5][];

	s_iNumObs = columns(s_mYg);
	s_iRestrict = 1;
	s_iPrint = 0;
			   		 
	s_vParmsOrig = 1;
	vParms = getStartingValues();
						
	logLikelihood(vParms, &dLoglike, &vGrad, &mHess);
	println(dLoglike);
	//return;			 
	
	MaxControl(-1,1,0);
	ir = MaxBFGS(logLikelihood, &vParms, &dLoglike, 0, 1);

	s_vParmsOrig = fabs(vParms);
	s_vParmsOrig = exp(round( 10* log(s_vParmsOrig) ) / 10 );
	vParms = 2 * (vParms .> 0) - 1;
	ir = MaxBFGS(logLikelihood, &vParms, &dLoglike, 0, 1);

	println("**********************************************************************************************");
	println("**********************************************************************************************");
	println("**********************************************************************************************");
	println("\n", MaxConvergenceMsg(ir),
          " using numerical derivatives",
          "\nFunction value = ", dLoglike);
	println("loglikelihood: ",dLoglike);	  

	println(restrictParms(vParms));
	
	vParms .*= s_vParmsOrig;
	s_vParmsOrig = 1;
	  

	//}
}