#include <oxstd.h>
#include <oxprob.h>
#include <oxdraw.h>
#include <oxfloat.h>
#import "maximize"
//#import <packages/maxsa/maxsa>

static decl s_mYc, s_mFactors;
static decl s_iNumObs, s_iNumFactors, s_iRestrict;
static decl s_iNumGrades, s_iNumFrailtyFactors, s_iNumMacros;
static decl s_vFactors, s_iPrint, s_iPlot;
static decl s_vParmsOrig;


restrictParms(const vParmsBegin)
{
	decl vParms = vParmsBegin .* s_vParmsOrig;
	decl vRestrictedParms = vParms;
	decl temp;
	decl inc = 0;

		// A parameters
		vRestrictedParms[inc : inc + s_iNumFactors - 1] = vParms[inc : inc + s_iNumFactors - 1];
		inc += s_iNumFactors;
		// B parameters
		vRestrictedParms[inc : inc + s_iNumFactors - 1] = 1 ./ (1 + exp(-vParms[inc : inc + s_iNumFactors - 1]));
		inc += s_iNumFactors;
		
		// vZc parameters
		inc += s_iNumGrades^2;
		// mZc
		//inc += (s_iNumGrades*s_iNumFactors) - s_iNumFrailtyFactors^2 - s_iNumGrades*s_iNumRecRateFactors;
		inc += s_iNumFrailtyFactors*s_iNumGrades - s_iNumFrailtyFactors*(s_iNumFrailtyFactors + 1)/2; 
	
	return vRestrictedParms;
}





getStartingValues()
{
	decl vParms = 0;

		s_iNumGrades = 4;
		s_iNumFrailtyFactors = 2;
	
		vParms = 0.0916 | 0.07525 |
		         3.40046 | 4.2775 |
				 6.299 | 8.288 | 9.626 | 10.293 |
				 -5.197 | 4.630 | 7.056 | 7.453 |
				 -7.636 | -5.455 | 4.648 | 5.974 |
				 -9.071 | -8.154 | -5.341 | 3.751 |
				0.0182419 | -0.456507 | 3.49525 | -0.668704 | -0.373708 ;
	
	s_iNumFactors = s_iNumFrailtyFactors;
	return vParms;
}





getGasMatrices(const vRestrictedParms, const pmA, const pmB, const pvZc, const pmZc, const iPrint)
{
	decl inc = 0, temp;
	
	pvZc[0] = 0;
	pmZc[0] = 0;
		

	 	pmA[0] = diag(vRestrictedParms[inc:inc + s_iNumFactors - 1]);
		inc += s_iNumFactors;
		pmB[0] = diag(vRestrictedParms[inc:inc + s_iNumFactors - 1]);
		inc += s_iNumFactors;
		
		pvZc[0] = reshape(vRestrictedParms[inc : inc + s_iNumGrades^2 -1], s_iNumGrades, s_iNumGrades);
		inc += s_iNumGrades^2;
		pmZc[0] = vRestrictedParms[inc++] ~ vRestrictedParms[inc++] |
		          1 ~ 0 |
				  vRestrictedParms[inc++] ~ vRestrictedParms[inc++] |
				  vRestrictedParms[inc++] ~ 1 ;

}





logLikelihood(const vParms, const pdLoglike, const pvG, const pmHess)
{

	decl mA, mB, mInfo, mInvInfo, mFactors, mYc;
	decl vZc, mZc;
	decl vIndex, vFactors, dLoglike, vScore, vSt, vRestrictedParms;
	decl iWarning, iNumObs, iNumGrades, iNumFactors;
	decl mEigval, mEigvec, mUU;
	decl t, i, j, k, sign;
	decl mPiTilde, mPi, mPiScore;
	pdLoglike[0] = -.Inf;					
			
	iNumObs = s_iNumObs;
	iNumFactors = s_iNumFactors;
	iNumGrades = s_iNumGrades;

	// load the data
	mYc = s_mYc;

	// transform parameters if s_iRestrict == 1
	if(s_iRestrict==1)
	{
		vRestrictedParms = restrictParms(vParms);
	}
	else
	{
		vRestrictedParms = vParms;
	}

	// Get the parameters of the transition and observation equation
	// y(t)|f(t) ~ p(. | vZc + mZc f(t)), f(t+1) = mB f(t) + mA score(t)
	getGasMatrices(vRestrictedParms, &mA, &mB, &vZc, &mZc, s_iPrint);

	// initialize the factors and scores
	mFactors = zeros(iNumFactors, iNumObs);
	vFactors = zeros(iNumFactors,1);
	dLoglike = 0;

	// loop over the observations to compute the GAS recursion and the likelihood
	for(t=0;t<iNumObs;t++)
	{

		// store the factors
		mFactors[][t] = vFactors;							

		// initialize the cumulative probabilities
		mPiTilde = 0 ~ (1.0 ./(1 + exp(-(vZc - mZc * vFactors)))) ~ 1;

		// compute the marginal probabilities for each transition
		mPi = mPiTilde[][1:] - mPiTilde[][:iNumGrades];
		mPi = (mPi .< 1e-6) .? 1e-6 .: mPi;
		mPiScore = mPiTilde[][1:] .* (1 - mPiTilde[][1:]) - mPiTilde[][:iNumGrades] .* (1 - mPiTilde[][:iNumGrades]);
			
		// compute log-likelihood increment
		dLoglike += sumc(mYc[][t] .* log(vecr(mPi)));
	
		// compute density score
		vScore = -sumc( (mYc[][t] ./ vecr(mPi) .* vecr(mPiScore)) .* (mZc ** ones(iNumGrades+1,1)) )';

		// compute density information matrix
		mInfo = ((mZc' ** ones(1,iNumGrades+1)) ) *
			(
				(sqr(vecr(mPiScore)) ./ vecr(mPi))
				.*
				(
					(mZc .* sumr(reshape(mYc[][t],iNumGrades,iNumGrades+1)))
					** ones(iNumGrades+1,1)
				)
			);


	 
		// check likelihood existence
		if(isnan(dLoglike))
		{
			break;
		}

		// update with GAS step
		if(isnan(mInfo) || (sumc(vecr(isdotinf(mInfo))) != 0))
		{
			println("Info Matrix FIX at step t=",t);
			mInvInfo = unit(s_iNumFactors);
		}
		else
		{
			eigensym(mInfo,&mEigval,&mEigvec);
		
			vIndex = <>;
			for(j=0;j<iNumFactors;j++)
			{
				if(mEigval[j]>0)
				{
					vIndex |= j;
				}
			}
			if(vIndex==<>)
			{
				mInvInfo = zeros(s_iNumFactors);
			}
			else
			{
				mUU = mEigvec[][vIndex];
				mInvInfo = mUU*(diag(1 ./sqrt(mEigval[vIndex])))*mUU';
			}
		}
		
		vSt = mInvInfo * vScore;
		
		vFactors = mB*vFactors + mA * vSt;
	}
					  
	s_vFactors = vFactors;
	s_mFactors = mFactors;

	pdLoglike[0] = double(dLoglike);
	
	if(isnan(pdLoglike[0]))
	{
		return 0;
	}
	else
	{
		return 1;
	}
}





computeGradient(const func, const vParms, const deltagrad)
{
	decl m, k, mI, mGradient, fbase, vParms2, vParmsTemp;
  	m = rows(vParms);
	k = columns(vParms);
	vParmsTemp = vParms;
	if(m==1)
	{
		vParmsTemp = vParms'; 
	}
	m = rows(vParmsTemp);
	mI = unit(m);
	fbase = func(vParmsTemp);
	mGradient = zeros(m,m);
	for(decl i=0;i<m;i++)
	{
		vParms2 = vParmsTemp + deltagrad*mI[][i];
		mGradient[][i] = (func(vParms2)-fbase)/deltagrad ;
	}
	return mGradient;
}





standardErrors(const objFunction, const restrictionFunction, const vParms, const pvParms, const pvStdErrors)
{
	decl mHess, mJacobian, ir;

	// ============================================ //
	//	compute standard errors
	// ============================================ //
	ir = Num2Derivative(objFunction, vParms, &mHess);
    if (!ir)
    {	print("Covar() failed in numerical second derivatives\n");
		pvParms[0] = zeros(vParms);
		pvStdErrors[0] = zeros(vParms); 
        return;
    }
    mJacobian = computeGradient(restrictionFunction,vParms, 0.001);
	pvParms[0] = restrictionFunction(vParms);
	pvStdErrors[0] = sqrt(diagonal(mJacobian*(invert(-mHess))*mJacobian'))';
}




simulateOrderLogitModel(const pmYc, const iNumObs, const vInitialFirms)
{
	decl mA, mB, mInfo, mInvInfo, mFactors, mYc, vYc;
	decl vZc, mZc;
	decl vIndex, vFactors, dLoglike, vScore, vSt, vRestrictedParms, vProbs;
	decl temp, iNumGrades, iNumFactors, iKK, iNumFirms, iNumDefaults;
	decl mEigval, mEigvec, mUU;
	decl t, i, j, k, sign;
	decl mPiTilde, mPi, mPiScore, mTrans;
				
	iNumFactors = s_iNumFactors;
	iNumGrades = s_iNumGrades;

	mYc = zeros(iNumGrades^2 + iNumGrades,iNumObs);
	iKK = rows(vInitialFirms) - 1;

	// simulation parameters
	mA = 0.091 ~ 0 |
	     0 ~ 0.075 ;
    mB = 0.98 ~ 0 |
	     0 ~ 0.985 ;
	vZc =  6.299 ~ 8.288 ~ 9.626 ~ 10.293 |
				 -5.197 ~ 4.630 ~ 7.056 ~ 7.453 |
				 -7.636 ~ -5.455 ~ 4.648 ~ 5.974 |
				 -9.071 ~ -8.154 ~ -5.341 ~ 3.751 ;
	mZc = 0.02 ~ -0.45 |
	      1 ~ 0 |
		  3.50 ~ -0.66 |
		  -0.40 ~ 1; 

	mFactors = zeros(iNumFactors, iNumObs);
	vFactors = zeros(iNumFactors,1);

	vYc = vInitialFirms;

	// Note: the stationary distribution of this model is for all firms to default
	// in the long-run unless new firms are added over time.
	mPiTilde = 0 ~ (1.0 ./(1 + exp(-vZc))) ~ 1;
	mPi = mPiTilde[][1:] - mPiTilde[][:iNumGrades];
	vProbs = sumc(mPi)/sumr(sumc(mPi));
	vProbs = vProbs[0:iKK-1]/sumr(vProbs[0:iKK-1]);
	
	for(t=0;t<iNumObs;t++)
	{
		
		// store the factors
		mFactors[][t] = vFactors;							

		mPiTilde = 0 ~ (1.0 ./(1 + exp(-(vZc - mZc * vFactors)))) ~ 1;
		
		// matrix of observed firm transitions
		mTrans = zeros(iKK,iKK+1);

		// Markov transition probability matrix for this time period
		mPi = mPiTilde[][1:] - mPiTilde[][:iNumGrades];
		mPi = (mPi .< 1e-6) .? 1e-6 .: mPi;
		mPiScore = mPiTilde[][1:] .* (1 - mPiTilde[][1:]) - mPiTilde[][:iNumGrades] .* (1 - mPiTilde[][:iNumGrades]);

		for(k=0;k<iKK;k++)
		{
			// number of firms in this ratings grade
			iNumFirms = vYc[k];

			// draw firm transitions
			mTrans[k][] = ranmultinomial(iNumFirms,mPi[k][]);
		}
	
		vYc = sumc(mTrans)';
		iNumDefaults = vYc[iKK]; // number of defaults
		mYc[][t] = vecr(mTrans);

		// We allow new firms to be added into the model. To keep the total number of firms the same
		// over time, we assume ``iNumDefaults'' firms are added for next time period.
		temp = ranmultinomial(iNumDefaults,vProbs)';
		vYc[0:iKK-1] = vYc[0:iKK-1] + temp; 
		vYc[iKK] = 0; 
	
		// compute density score
		vScore = -sumc( (mYc[][t] ./ vecr(mPi) .* vecr(mPiScore)) .* (mZc ** ones(iNumGrades+1,1)) )';

		// compute density information matrix
		mInfo = ((mZc' ** ones(1,iNumGrades+1)) ) *
			(
				(sqr(vecr(mPiScore)) ./ vecr(mPi))
				.*
				(
					(mZc .* sumr(reshape(mYc[][t],iNumGrades,iNumGrades+1)))
					** ones(iNumGrades+1,1)
				)
			);
	
		// update with GAS step
		if(isnan(mInfo) || (sumc(vecr(isdotinf(mInfo))) != 0))
		{
			mInvInfo = unit(s_iNumFactors);
		}
		else
		{
			eigensym(mInfo,&mEigval,&mEigvec);
		
			vIndex = <>;
			for(j=0;j<iNumFactors;j++)
			{
				if(mEigval[j]>0)
				{
					vIndex |= j;
				}
			}
			if(vIndex==<>)
			{
				mInvInfo = zeros(s_iNumFactors);
			}
			else
			{
				mUU = mEigvec[][vIndex];
				mInvInfo = mUU*(diag(1 ./sqrt(mEigval[vIndex])))*mUU';
			}
		}
		
		vSt = mInvInfo * vScore;
		
		vFactors = mB*vFactors + mA * vSt;
	}
	pmYc[0] = mYc;	
}

main()
{

	decl vParms, vParmsMLE, vStdErrors;
	decl dLoglike, ir, i, j, t, iNumTotalObs;
	decl mYc;
	decl vInitialFirms;

	// initialize the data formats and output options
	s_iNumGrades = 4;			// number of rating buckets (exclusive default)
	s_iNumObs = 750;			// number of observations (simulated data)
	s_iRestrict = 1;
	s_iPrint = 1;				// put intermediate printing of results on(1)/off(0)
	s_iPlot = 0;				// put intermediate plots on(1)/off(0)
	s_iNumFactors = 2;			// number of frailty factors

	// set the initial number of firms over the s_iNumGrades categories
	vInitialFirms = <1146; 263; 613; 311; 0>;

	// generate data and store simulated data in global variable s_mYc
	simulateOrderLogitModel(&mYc, s_iNumObs, vInitialFirms);
//	s_iNumObs -= 5000; mYc = mYc[][5000:];
	s_mYc = mYc;
//	for (i=0;i<s_iNumGrades;++i) for (j=0;j<=s_iNumGrades;++j)
//	{
//		println(i*(s_iNumGrades+1)+j);
//		Draw(i*(s_iNumGrades+1)+j,mYc[i*(s_iNumGrades+1)+j][] ./ sumc(mYc[i*(s_iNumGrades+1):i*(s_iNumGrades+1)+s_iNumGrades][]));
//	}
//	DrawAdjust(ADJ_AREAMATRIX,s_iNumGrades,s_iNumGrades+1); SetDraw(SET_COLORMODEL,3);
//	SaveDrawWindow("test.eps"); ShowDrawWindow(); exit(0);

	// initialize the parameters
	s_vParmsOrig = 1;
	vParms = getStartingValues();

	// compute the initial likelihood
	decl vG, mHess;
	logLikelihood(vParms, &dLoglike, &vG, &mHess);
	

	// maximize the likelihood in two steps; first for unscaled parameters;
	// then for scaled parameters to get better numerical properties
	// (use your own optimizer if you have one)
	// STEP 1:
	MaxControl(1000,1,0);
	ir = MaxBFGS(logLikelihood, &vParms, &dLoglike, 0, 1);

	// STEP 2: s_vParmsOrig roughly are rounded versions of the parameter estimates
	// then we estimate multiples of this using a second optimization round
	s_vParmsOrig = fabs(vParms);
	s_vParmsOrig = exp(round( 10* log(s_vParmsOrig) ) / 10 );
	vParms = 2 * (vParms .> 0) - 1;
	MaxControl(1000,1,0);
	ir = MaxBFGS(logLikelihood, &vParms, &dLoglike, 0, 1);

	// Print the output of the estimation
	println("**********************************************************************************************");
	println("**********************************************************************************************");
	println("**********************************************************************************************");
	println("\n", MaxConvergenceMsg(ir),
          " using numerical derivatives",
          "\nFunction value = ", dLoglike);
	println("loglikelihood: ",dLoglike);	  
	
	println(restrictParms(vParms));
	
	vParms .*= s_vParmsOrig;
	s_vParmsOrig = 1;
	  

	println(" ");
	for(decl j=0;j<rows(vParms); j++)
		print(vParms[j]," | ");
	println(" ");
				
	standardErrors(logLikelihood, restrictParms, vParms, &vParmsMLE, &vStdErrors);
					  
	s_iPrint = 1;
	s_iRestrict = 0;
	logLikelihood(vParmsMLE,&dLoglike,0,0);
	logLikelihood(vStdErrors,&dLoglike,0,0);

	println(vParmsMLE);
	println(vStdErrors);

	//}
}